
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "read_hit_heap.h"


static inline int32_t hash_e(read_hit_heap_e * e) {
	return 0;
}

static inline int32_t hash_find(read_hit_heap * rhh,read_hit_heap_e * e) {
	int32_t current_slot=rhh->ll[hash_e(e)];
	while (current_slot!=-1) {
		read_hit_heap_e * other=rhh->array+current_slot;
		if (e->rh[0]->g_off==other->rh[0]->g_off &&
			e->rh[0]->cn==other->rh[0]->cn &&
			e->rh[0]->st==other->rh[0]->st) {
			if (e->rh[1]==NULL && other->rh[1]==NULL) {
				return current_slot;
			} else if (e->rh[1]!=NULL && other->rh[1]!=NULL) {
				if (e->rh[1]->g_off==other->rh[1]->g_off &&
					e->rh[1]->cn==other->rh[1]->cn &&
					e->rh[1]->st==other->rh[1]->st) {
					return current_slot;
				}
			}
		}	
		current_slot=other->next;
	}
	return current_slot;
} 

static inline void hash_it(read_hit_heap * rhh, read_hit_heap_e * e) {
	e->hash=hash_e(e);
	if (rhh->ll[e->hash]==-1) {
		rhh->ll[e->hash]=e->slot;
		e->next=-1;
		e->prev=-1;
	} else {
		e->next=rhh->ll[e->hash];
		rhh->ll[e->hash]=e->slot;
		e->prev=-1;
	}
}


static inline bool rhh_compare(read_hit_heap_e * a, read_hit_heap_e * b) {
	if (a->score>=b->score) {
		return false;
	} else if (a->score==b->score) {
		if (a->isize_score<=b->isize_score) {
			return false;
		} else {
			return true;
		}
	}
	return true;
}


void read_hit_heap_init(read_hit_heap * rhh, uint32_t capacity,int32_t hash_size) {
	assert(capacity>0);
	assert(rhh != NULL);
	rhh->array = (struct read_hit_heap_e *)malloc(capacity * sizeof(struct read_hit_heap_e)); 
	if (rhh->array==NULL) { 
		fprintf(stderr,"Failed to allocate heap!\n"); 
		exit(1); 
	} 
	rhh->slot_array = (int32_t *)malloc(capacity * sizeof(int32_t)); 
	if (rhh->slot_array==NULL) { 
		fprintf(stderr,"Failed to allocate heap!\n"); 
		exit(1); 
	}
	rhh->ll=(int32_t*)malloc(hash_size * sizeof(int32_t));
	if (rhh->ll==NULL) {
		fprintf(stderr,"Failed to allocate ll\n");
		exit(1);
	}
	int32_t i;
	for (i=0; i<hash_size; i++) {
		rhh->ll[i]=-1;
	}
	rhh->hash_size=hash_size;
	rhh->capacity = capacity;
	rhh->load = 0;
}

void read_hit_heap_destroy(struct read_hit_heap * rhh) {
	assert(rhh != NULL);
	free(rhh->array);
	free(rhh->slot_array);
}

static inline void read_hit_heap_percolate_up(struct read_hit_heap * rhh, uint32_t node) {
	int32_t tmp;
	uint32_t parent;
	parent = node / 2;
	while (node > 1 && rhh_compare(rhh->array+rhh->slot_array[node-1],rhh->array+rhh->slot_array[parent-1])) {
		tmp = rhh->slot_array[parent-1];
		rhh->slot_array[parent-1] = rhh->slot_array[node-1];
		rhh->slot_array[node-1] = tmp;
		node = parent;
		parent = node / 2;
	}
}

static inline void read_hit_heap_percolate_down(struct read_hit_heap * rhh, int32_t node) {
	int32_t tmp;
	int32_t left, right, max;
	do {
		left = node * 2;
		right = left + 1;
		max = node;
		if (left <= rhh->load && rhh_compare(rhh->array+rhh->slot_array[left-1],rhh->array+rhh->slot_array[node-1])) 
		max = left;
		if (right <= rhh->load && rhh_compare(rhh->array+rhh->slot_array[right-1],rhh->array+rhh->slot_array[max-1])) 
		max = right;
		if (max == node)
		break;
		tmp = rhh->slot_array[max-1];
		rhh->slot_array[max-1] = rhh->slot_array[node-1];
		rhh->slot_array[node-1] = tmp;
		node = max;
	} while (1);
}

/*void read_hit_heap_extract_min(struct read_hit_heap * rhh, struct read_hit_heap_e * dest)  {
	assert(rhh != NULL && rhh->load > 0);
	*dest = rhh->array[0];
	rhh->load--;
	if (rhh->load > 0) {
		rhh->array[0] = rhh->array[rhh->load];
		read_hit_heap_percolate_down(rhh, 1);
	}
}*/

void read_hit_heap_replace_min(struct read_hit_heap * rhh, struct read_hit_heap_e * e) {
	assert(rhh != NULL && rhh->load > 0);
	int32_t previous_node = rhh->array[rhh->slot_array[0]].prev;
	if (previous_node>=0) {
		rhh->array[previous_node].next=rhh->array[rhh->slot_array[0]].next;
	} else {
		rhh->ll[rhh->array[rhh->slot_array[0]].hash]=rhh->array[rhh->slot_array[0]].next;
	}
	e->slot=rhh->slot_array[0];
	e->prev=-1;
	e->next=-1;
	rhh->array[rhh->slot_array[0]] = *e;
	hash_it(rhh,rhh->array+0);
	read_hit_heap_percolate_down(rhh, 1);
}

/*void read_hit_heap_get_min(struct read_hit_heap * rhh, struct read_hit_heap_e * dest) {
	assert(rhh != NULL && rhh->load > 0);
	*dest = rhh->array[0];
}*/

void read_hit_heap_insert(struct read_hit_heap * rhh, struct read_hit_heap_e * e) {
	assert(rhh != NULL && rhh->load < rhh->capacity);
	e->slot=rhh->load;
	e->prev=-1;
	e->next=-1;
	rhh->array[rhh->load] = *e;
	rhh->slot_array[rhh->load] = rhh->load;
	hash_it(rhh,rhh->array+rhh->load);
	rhh->load++;
	read_hit_heap_percolate_up(rhh, rhh->load);
}

void read_hit_heap_insert_bounded(struct read_hit_heap * rhh, struct read_hit_heap_e *e) {
	if (rhh->load<rhh->capacity) {
		read_hit_heap_insert(rhh,e);
	} else if (rhh_compare(rhh->array+rhh->slot_array[0],e)) {
		assert(rhh->load==rhh->capacity);
		read_hit_heap_replace_min(rhh,e);
	}
}

void read_hit_heap_insert_bounded_strata(struct read_hit_heap * rhh, struct read_hit_heap_e *e) {
	if (rhh->load==0) {
		read_hit_heap_insert(rhh,e);
	} else if (rhh_compare(e,rhh->array+rhh->slot_array[0])) {
		return;
	} else if (rhh_compare(rhh->array+rhh->slot_array[0],e)) {
		//means that we should wipe heap and take the new one
		int32_t i;
		for (i=0; i<rhh->hash_size; i++) {
			rhh->ll[i]=-1;
		}
		rhh->load=0;
		read_hit_heap_insert(rhh,e);
	} else if (rhh->load<rhh->capacity) {
		//equal
		read_hit_heap_insert(rhh,e);
	}
}


void read_hit_heap_heapify(struct read_hit_heap * rhh){
	assert(rhh != NULL);
	uint32_t node;
	for (node = rhh->load / 2; node >= 1; node--) {
		read_hit_heap_percolate_down(rhh, node);
	}
}


